﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.IO;
using System.Runtime.InteropServices;
using System.Net.Sockets;
using System.Threading;

namespace AutoUTorrent
{
    public partial class Form1 : Form
    {
        protected const string multicastHeader = "Autoutorrent:";
        protected string sessionId;
        private DateTime otherMachinesBusyAt = DateTime.FromFileTime(0);
        private const string multicastBusySignal = "busy";
        private volatile bool throttled = false;
        private volatile bool isIdle = false;
        private bool configChanged = false;
        private CookieContainer cookieJar = new CookieContainer();
        private MulticastListener multicastListener;


        struct LastInputInfo
        {
            public uint cbSize;
            public uint dwTime;
        }


        [DllImport("User32.dll")]
        private static extern bool GetLastInputInfo(ref LastInputInfo plii);


        public Form1()
        {
            InitializeComponent();
            GenerateSessionId();
        }


        // generates a random string we can use in multicasting to distinguish this instance
        // useful for testing multiple instances on a single PC
        private void GenerateSessionId()
        {
            Random rnd = new Random();
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < 10; i++)
            {
                sb.Append((char)rnd.Next(65) + 32);
            }
            sessionId = sb.ToString();
        }


        void SendRequestToUTorrent(int uploadSpeed, int downloadSpeed)
        {
            this.txtActivity.AppendText("Setting upload speed " + uploadSpeed + ", download speed " + downloadSpeed + "\r\n");
            string token = GetToken();
            string bandwidthResponse = MakeWebRequest(txtWebUIAddress.Text + "?token=" + token +
                "&action=setsetting&s=max_ul_rate&v=" + uploadSpeed +
                "&s=max_dl_rate&v=" + downloadSpeed);
        }


        string GetToken()
        {
            string tokenRequestDetails = MakeWebRequest(txtWebUIAddress.Text + "token.html");
            int start = tokenRequestDetails.IndexOf(">", 9);
            int end = tokenRequestDetails.IndexOf("<", start + 1);
            string token = tokenRequestDetails.Substring(start + 1, end - start - 1);
            this.txtToken.Text = token;
            return token;
        }


        string MakeWebRequest(string url)
        {
            this.txtUrl.Text = url;
            this.txtResponse.Text = "";
            StringBuilder sb = new StringBuilder();
            byte[] buf = new byte[8192];
            HttpWebRequest request = (HttpWebRequest)WebRequest.Create(url);
            request.CookieContainer = cookieJar;
            string authInfo = this.txtUser.Text + ":" + this.txtPassword.Text;
            authInfo = Convert.ToBase64String(Encoding.Default.GetBytes(authInfo));
            request.Headers["Authorization"] = "Basic " + authInfo;
            HttpWebResponse response = (HttpWebResponse)request.GetResponse();
            Stream resStream = response.GetResponseStream();
            string tempString = null;
            int count = 0;
            do
            {
                count = resStream.Read(buf, 0, buf.Length);
                if (count != 0)
                {
                    tempString = Encoding.ASCII.GetString(buf, 0, count);
                    sb.Append(tempString);
                }
            }
            while (count > 0);
            string fullResponseText = sb.ToString();
            this.txtResponse.Text = fullResponseText;
            foreach (Cookie cookie in response.Cookies)
            {
                if (cookie.Name == "GUID")
                {
                    this.txtGuidCookie.Text = cookie.Value;
                }
            }
            return fullResponseText;
        }



        public static uint GetIdleTime()
        {
            LastInputInfo lastInput = new LastInputInfo();
            lastInput.cbSize = (uint)System.Runtime.InteropServices.Marshal.SizeOf(lastInput);
            GetLastInputInfo(ref lastInput);
            return ((uint)Environment.TickCount - lastInput.dwTime);
        }


        private bool areOtherMachinesIdle()
        {
            bool areOtherMachinesIdle = (DateTime.Now.Subtract(this.otherMachinesBusyAt).TotalSeconds > tbIdleTimeout.Value);
            return areOtherMachinesIdle;
        }


        private void timer1_Tick(object sender, EventArgs e)
        {
            try
            {
                if (txtActivity.Lines.Length > 100)
                {
                    txtActivity.Clear();
                }

                // update our status
                int idleSecs = (int)(GetIdleTime() / 1000);
                this.lblIdleFor.Text = String.Format("{0}", idleSecs);
                if (idleSecs > tbIdleTimeout.Value && !isIdle)
                {
                    txtActivity.AppendText("This PC is now idle\r\n");
                    this.isIdle = true;
                }
                else if (idleSecs < tbIdleTimeout.Value && isIdle)
                {
                    txtActivity.AppendText("This PC is now busy\r\n");
                    isIdle = false;
                }

                if (throttled && this.isIdle && areOtherMachinesIdle())
                {
                    txtActivity.AppendText("All PCs are now idle, speeding up uTorrent\r\n");
                    btnSendIdle_Click(null, null);
                }
                else if (!throttled && (!this.isIdle || !areOtherMachinesIdle()))
                {
                    txtActivity.AppendText("At least one PC is busy, throttling uTorrent\r\n");
                    btnSendBusy_Click(null, null);
                }
                if (cbMulticastEnabled.Checked)
                {
                    sendMulticast();
                }
            }
            catch (Exception ex)
            {
                this.lblBandwidthStatus.Text = "Error";
                this.lblBandwidthStatus.ForeColor = Color.Red;
                this.txtResponse.Text = ex.ToString();
                this.txtActivity.AppendText(ex.ToString() + "\r\n");
            }
        }


        private void btnSendIdle_Click(object sender, EventArgs e)
        {
            try
            {
                SendRequestToUTorrent(int.Parse(this.txtIdleUploadSpeed.Text), int.Parse(this.txtIdleDownloadSpeed.Text));
                throttled = false;
                this.lblBandwidthStatus.Text = "Fast";
                this.lblBandwidthStatus.ForeColor = Color.Green;
                if (cbShowNotifications.Checked && this.WindowState == FormWindowState.Minimized)
                {
                    notifyIcon1.ShowBalloonTip(500, "AutoUTorrent", "PC idle, UTorrent speeded up", ToolTipIcon.Info);
                }
            }
            catch (Exception ex)
            {
                this.txtResponse.Text = ex.Message;
                this.txtActivity.AppendText(ex.Message + "\r\n");
                if (cbShowNotifications.Checked && this.WindowState == FormWindowState.Minimized)
                {
                    notifyIcon1.ShowBalloonTip(500, "AutoUTorrent", "Error speeding up UTorrent: " + ex.Message, ToolTipIcon.Info);
                }
            }
        }


        private void btnSendBusy_Click(object sender, EventArgs e)
        {
            try
            {
                SendRequestToUTorrent(int.Parse(this.txtBusyUploadSpeed.Text), int.Parse(this.txtBusyDownloadSpeed.Text));
                throttled = true;
                this.lblBandwidthStatus.Text = "Throttled";
                this.lblBandwidthStatus.ForeColor = Color.Orange;
                if (cbShowNotifications.Checked && this.WindowState == FormWindowState.Minimized)
                {
                    notifyIcon1.ShowBalloonTip(500, "AutoUTorrent", "PC not idle, UTorrent speed throttled", ToolTipIcon.Info);
                }
            }
            catch (Exception ex)
            {
                this.txtResponse.Text = ex.Message;
                this.txtActivity.AppendText(ex.Message + "\r\n");
                if (cbShowNotifications.Checked && this.WindowState == FormWindowState.Minimized)
                {
                    notifyIcon1.ShowBalloonTip(500, "AutoUTorrent", "Error throttling UTorrent: " + ex.Message, ToolTipIcon.Info);
                }
            }
        }


        private void cbAutoMode_CheckedChanged(object sender, EventArgs e)
        {
            this.timer1.Enabled = this.cbAutoMode.Checked;
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (configChanged)
            {
                if (MessageBox.Show("Save config?", "AutoUTorrent", MessageBoxButtons.YesNo) == DialogResult.Yes)
                {
                    SaveConfig();
                }
            }
        }


        string getConfigFileLocation()
        {
            string filename = Path.GetDirectoryName(Application.ExecutablePath) + "\\AutoUTorrent.conf";
            return filename;
        }


        private void LoadConfig()
        {
            try
            {
                IniFile ini = new IniFile(getConfigFileLocation());
                this.txtUser.Text = ini.IniReadValue("UI", "User", "admin");
                this.txtPassword.Text = ini.IniReadValue("UI", "Password", "password");
                this.txtWebUIAddress.Text = ini.IniReadValue("UI", "URL", "http://localhost:18424/gui/");
                this.txtIdleUploadSpeed.Text = ini.IniReadValue("Idle", "UploadSpeed", "0");
                this.txtIdleDownloadSpeed.Text = ini.IniReadValue("Idle", "DownloadSpeed", "0");
                this.txtBusyUploadSpeed.Text = ini.IniReadValue("Busy", "UploadSpeed", "10");
                this.txtBusyDownloadSpeed.Text = ini.IniReadValue("Busy", "DownloadSpeed", "25");
                this.tbCheckInterval.Value = int.Parse(ini.IniReadValue("Timer", "CheckInterval", "1"));
                this.lblCheckIntervalValue.Text = "" + tbCheckInterval.Value;
                this.timer1.Interval = tbCheckInterval.Value * 1000;
                this.tbIdleTimeout.Value = int.Parse(ini.IniReadValue("Timer", "IdleTimeout", "10"));
                this.lblIdleTimeoutValue.Text = "" + tbIdleTimeout.Value;
                this.cbShowNotifications.Checked = ini.IniReadValue("Misc", "ShowNotifications", "true") == "true";
                this.cbMulticastEnabled.Checked = ini.IniReadValue("Multicast", "Enabled", "true") == "true";
                this.txtMulticastIp.Text = ini.IniReadValue("Multicast", "Ip", "239.0.0.222");
                this.txtMulticastPort.Text = ini.IniReadValue("Multicast", "Port", "2222");
                if (cbMulticastEnabled.Checked)
                {
                    txtActivity.AppendText("Multicasting enabled on " + txtMulticastIp.Text + ":" + txtMulticastPort.Text + "\r\n");
                }
                else
                {
                    txtActivity.AppendText("Multicasting disabled\r\n");
                }
                configHasChanged(false);
            }
            catch (Exception e)
            {
                txtActivity.AppendText(e.ToString() + "\r\n");
            }
        }

        private void SaveConfig()
        {
            IniFile ini = new IniFile(getConfigFileLocation());
            ini.IniWriteValue("UI", "User", this.txtUser.Text);
            ini.IniWriteValue("UI", "Password", this.txtPassword.Text);
            ini.IniWriteValue("UI", "URL", this.txtWebUIAddress.Text);
            ini.IniWriteValue("Idle", "UploadSpeed", this.txtIdleUploadSpeed.Text);
            ini.IniWriteValue("Idle", "DownloadSpeed", this.txtIdleDownloadSpeed.Text);
            ini.IniWriteValue("Busy", "UploadSpeed", this.txtBusyUploadSpeed.Text);
            ini.IniWriteValue("Busy", "DownloadSpeed", this.txtBusyDownloadSpeed.Text);
            ini.IniWriteValue("Timer", "CheckInterval", "" + this.tbCheckInterval.Value);
            ini.IniWriteValue("Timer", "IdleTimeout", "" + this.tbIdleTimeout.Value);
            ini.IniWriteValue("Misc", "ShowNotifications", cbShowNotifications.Checked ? "true" : "false");
            ini.IniWriteValue("Multicast", "Enabled", this.cbMulticastEnabled.Checked ? "true" : "false");
            ini.IniWriteValue("Multicast", "Ip", this.txtMulticastIp.Text);
            ini.IniWriteValue("Multicast", "Port", "" + this.txtMulticastPort.Text);
            configHasChanged(false);
        }


        private void configHasChanged(bool hasChanged)
        {
            configChanged = hasChanged;
            this.btnSaveConfig.Enabled = hasChanged;
        }


        private void txtIdleUploadSpeed_TextChanged(object sender, EventArgs e)
        {
            configHasChanged(true);
        }

        private void txtIdleDownloadSpeed_TextChanged(object sender, EventArgs e)
        {
            configHasChanged(true);
        }

        private void txtBusyUploadSpeed_TextChanged(object sender, EventArgs e)
        {
            configHasChanged(true);
        }

        private void txtBusyDownloadSpeed_TextChanged(object sender, EventArgs e)
        {
            configHasChanged(true);
        }

        private void txtUser_TextChanged(object sender, EventArgs e)
        {
            configHasChanged(true);
        }

        private void txtPassword_TextChanged(object sender, EventArgs e)
        {
            configHasChanged(true);
        }


        private void txtWebUIAddress_TextChanged(object sender, EventArgs e)
        {
            configHasChanged(true);
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            LoadConfig();
            
            SetTooltip(this.txtWebUIAddress, "Enter the base URL that UTorrent's WebUI is listening on.  Set this in UTorrent's preferences under the Web UI section");
            SetTooltip(this.lblWebAddressDescription, "Enter the base URL that UTorrent's WebUI is listening on.  Set this in UTorrent's preferences under the Web UI section");
            SetTooltip(this.txtUser, "The user ID to authenticate against UTorrent with.  Set this in UTorrent's preferences under the Web UI section");
            SetTooltip(this.lblUserDescription, "The user ID to authenticate against UTorrent with.  Set this in UTorrent's preferences under the Web UI section");
            SetTooltip(this.txtPassword, "The password to authenticate against UTorrent with.  Set this in UTorrent's preferences under the Web UI section");
            SetTooltip(this.lblPasswordDescription, "The password to authenticate against UTorrent with.  Set this in UTorrent's preferences under the Web UI section");
            SetTooltip(this.txtIdleDownloadSpeed, "UTorrent's download speed will be set to this value when your PC is idle");
            SetTooltip(this.txtIdleUploadSpeed, "UTorrent's upload speed will be set to this value when your PC is idle");
            SetTooltip(this.txtBusyDownloadSpeed, "UTorrent's download speed will be set to this value when your PC is busy");
            SetTooltip(this.txtBusyUploadSpeed, "UTorrent's upload speed will be set to this value when your PC is busy");
            SetTooltip(this.tbCheckInterval, "How often (seconds) AutoUTorrent should check to see if your PC is idle.  Around 5 is a good value");
            SetTooltip(this.lblCheckIntervalDescription, "How often (seconds) AutoUTorrent should check to see if your PC is idle.  Around 5 is a good value");
            SetTooltip(this.lblCheckIntervalValue, "How often (seconds) AutoUTorrent should check to see if your PC is idle.  Around 5 is a good value");
            SetTooltip(this.btnSendBusy, "Send a command to UTorrent to set the bandwidth speeds to the 'Busy' values");
            SetTooltip(this.btnSendIdle, "Send a command to UTorrent to set the bandwidth speeds to the 'Idle' values");
            SetTooltip(this.cbAutoMode, "Switches the idle-checking on/off (useful for testing).  Usually you should have this box checked");
            SetTooltip(this.lblIdleForLabel, "This shows how many seconds your PC has been idle");
            SetTooltip(this.lblIdleFor, "This shows how many seconds your PC has been idle");
            SetTooltip(this.tbIdleTimeout, "AutoUTorrent will send the Idle bandwidth settings to UTorrent when your PC has been idle for at least this number of seconds");
            SetTooltip(this.lblIdleTimeoutDescription, "AutoUTorrent will send the Idle bandwidth settings to UTorrent when your PC has been idle for at least this number of seconds");
            SetTooltip(this.lblIdleTimeoutValue, "AutoUTorrent will send the Idle bandwidth settings to UTorrent when your PC has been idle for at least this number of seconds");
            SetTooltip(this.btnSaveConfig, "Saves config to 'AutoUTorrent.conf'");
            SetTooltip(this.cbMulticastEnabled, "Broadcast our idle status to other AutoUTorrent instances on this network");
            SetTooltip(this.txtMulticastIp, "The multicast ip address to use.  All AutoUTorrent instances on this network must share the same address");
            SetTooltip(this.txtMulticastPort, "The multicast port to use.  All AutoUTorrent instances on this network must share the same port");
            SetTooltip(this.cbShowNotifications, "Show system tray notifications when application is minimised");
        }


        void SetTooltip(Control c, string text)
        {
            ToolTip ToolTip1 = new ToolTip();
            ToolTip1.SetToolTip(c, text);
        }

        private void checkInterval_Scroll(object sender, EventArgs e)
        {
            configHasChanged(true);
            lblCheckIntervalValue.Text = "" + tbCheckInterval.Value;
            this.timer1.Interval = tbCheckInterval.Value * 1000;
        }

        private void btnSaveConfig_Click(object sender, EventArgs e)
        {
            SaveConfig();
        }

        private void IdleTimeout_Scroll(object sender, EventArgs e)
        {
            configHasChanged(true);
            lblIdleTimeoutValue.Text = "" + tbIdleTimeout.Value;
        }


        private void Form1_Resize(object sender, EventArgs e)
        {
            // minimise to system tray
            if (FormWindowState.Minimized == this.WindowState)
            {
                this.notifyIcon1.Visible = true;
                if (!cbAutoMode.Checked)
                {
                    notifyIcon1.ShowBalloonTip(500, "AutoUTorrent", "Warning - Auto Mode is disabled.  Are you sure this is intentional!?", ToolTipIcon.Info);
                }
                else
                {
                    if (cbShowNotifications.Checked)
                    {
                        notifyIcon1.ShowBalloonTip(500, "AutoUTorrent", "AutoUTorrent is running in the background", ToolTipIcon.Info);
                    }
                }
                this.Hide();
            }
            else if (FormWindowState.Normal == this.WindowState)
            {
                notifyIcon1.Visible = false;
            }
        }

        private void notifyIcon1_Click(object sender, EventArgs e)
        {
            this.Show();
            this.WindowState = FormWindowState.Normal;
        }

        private void cbShowNotifications_CheckedChanged(object sender, EventArgs e)
        {
            configHasChanged(true);
        }


        private void startListeningForMulticasts()
        {
            multicastListener = new MulticastListener(this, this.txtMulticastIp.Text, int.Parse(this.txtMulticastPort.Text));
            multicastListener.startThread();
        }


        private void stopListeningForMulticasts()
        {
            if (multicastListener != null)
            {
                multicastListener.stopThread();
                multicastListener = null;
                sendMulticast(); // this makes sure the listening thread picks up a message & re-evaluates the running flag
            }
        }


        private void sendMulticast()
        {
            using (UdpClient udpclientSend = new UdpClient())
            {
                try
                {
                    IPAddress multicastaddress = IPAddress.Parse(txtMulticastIp.Text);
                    udpclientSend.JoinMulticastGroup(multicastaddress);
                    IPEndPoint remoteep = new IPEndPoint(multicastaddress, int.Parse(txtMulticastPort.Text));
                    string idleStatusAsString = (this.isIdle ? "idle" : multicastBusySignal);
                    this.txtActivity.InvokeEx(c => this.txtActivity.AppendText("Multicasting status '" + idleStatusAsString + "'\r\n"));
                    Byte[] buffer = Encoding.Unicode.GetBytes(multicastHeader + "," + sessionId + "," + Environment.MachineName + "," + idleStatusAsString);
                    udpclientSend.Send(buffer, buffer.Length, remoteep);
                    udpclientSend.DropMulticastGroup(multicastaddress);
                }
                catch (Exception e)
                {
                    this.txtActivity.InvokeEx(c => this.txtActivity.AppendText(e.ToString() + "\r\n"));
                }
                finally
                {
                    udpclientSend.Close();
                }
            }
        }

        private void cbMulticastEnabled_CheckedChanged(object sender, EventArgs e)
        {
            configHasChanged(true);
            if (cbMulticastEnabled.Checked)
            {
                startListeningForMulticasts();
            }
            else
            {
                stopListeningForMulticasts();
            }
        }

        private void txtMulticastIp_TextChanged(object sender, EventArgs e)
        {
            configHasChanged(true);
        }

        private void txtMulticastPort_TextChanged(object sender, EventArgs e)
        {
            configHasChanged(true);
        }



        class MulticastListener
        {
            private readonly Form1 parent;
            private readonly string multicastIp;
            private readonly int multicastPort;
            private volatile bool shouldRun;

            public MulticastListener(Form1 parent, string multicastIp, int multicastPort)
            {
                this.parent = parent;
                this.multicastIp = multicastIp;
                this.multicastPort = multicastPort;
                this.shouldRun = true;
            }


            public void startThread()
            {
                Thread listener = new Thread(() =>
                {
                    this.parent.txtActivity.InvokeEx(c => this.parent.txtActivity.AppendText(String.Format("Listening for multicasts on {0}:{1}\r\n", multicastIp, multicastPort)));
                    while (shouldRun)
                    {
                        using (UdpClient udpclientListen = new UdpClient())
                        {
                            try
                            {
                                IPEndPoint localEp = new IPEndPoint(IPAddress.Any, multicastPort);
                                udpclientListen.Client.Bind(localEp);
                                IPAddress multicastaddress = IPAddress.Parse(multicastIp);
                                udpclientListen.JoinMulticastGroup(multicastaddress, 255);                                
                                Byte[] data = udpclientListen.Receive(ref localEp);
                                string strData = Encoding.Unicode.GetString(data);
                                // this.parent.txtActivity.InvokeEx(c => this.parent.txtActivity.AppendText(String.Format("Debug: Received multicast '{0}''\r\n", strData)));
                                if (strData.StartsWith(multicastHeader)) // check this is from an AutoUTorrent instance and not some other application
                                {
                                    string[] strDataSections = strData.Split(',');
                                    string otherPCsessionId = strDataSections[1];
                                    string otherPCname = strDataSections[2];
                                    string otherPCstatus = strDataSections[3];
                                    if (this.parent.sessionId != otherPCsessionId)
                                    {
                                        this.parent.txtActivity.InvokeEx(c => this.parent.txtActivity.AppendText(String.Format("Received multicast from '{0}' session '{1}' status '{2}'\r\n", otherPCname, otherPCsessionId, otherPCstatus)));
                                        if (otherPCstatus == multicastBusySignal)
                                        {
                                            // another PC was active at this time...the main timer will pick up this flag
                                            this.parent.otherMachinesBusyAt = DateTime.Now;
                                        }
                                    }
                                }
                                else
                                {
                                    this.parent.txtActivity.InvokeEx(c => this.parent.txtActivity.AppendText("Received unrecognised multicast: '" + strData + "'\r\n"));
                                }
                                udpclientListen.DropMulticastGroup(multicastaddress);
                            }
                            catch (Exception se)
                            {
                                this.parent.txtActivity.InvokeEx(c => this.parent.txtActivity.AppendText("Received exception: '" + se.Message + "'\r\n"));
                                Thread.Sleep(5000);
                            }
                            finally
                            {
                                udpclientListen.Close();
                            }
                        }
                    }
                    this.parent.txtActivity.InvokeEx(c => this.parent.txtActivity.AppendText(String.Format("Stopped listening for multicasts on {0}:{1}\r\n", multicastIp, multicastPort)));
                });
                listener.IsBackground = true;
                listener.Start();
            }


            public void stopThread()
            {
                this.shouldRun = false;
            }

        }


    }
}
